function [x,c] = linffitting_solve(K,yd,alpha,xe,opts)
% [x,c] = LINFFITTING_SOLVE(K,yd,alpha,xe,opts)
% semi-smooth Newton method for L^\inf-fitting-L^2-regularization using a
%   path-following method
% Input:  K      -- operator to be inverted
%         yd     -- given data
%         alpha  -- regularization parameter for primal problem
%         xe     -- exact solution for plotting)
%         opts   -- options for semismooth Newton method (see below)
% Output: x      -- Linf reconstruction
%         c      -- Linf bound norm of residual
%
% Christian Clason(christian.clason@uni-graz.at)
% February 9, 2012

%% Set parameters
N = size(K,2);

% Default parameters
if nargin <= 4 || isempty(opts)
    maxit   = 20;              % maximum number of iterations
    dispopt = 0;               % plot iterates in SSN iteration
else
    maxit   = opts.maxit;      % maximum number of iterations
    dispopt = opts.dispopt;    % plot iterates in SSN iteration
end

% precompute parts of Newton derivative
Kt = K';
aI = alpha*speye(N);

% initialize iterates
ap = zeros(N,1); ap_old = ap; % active set for upper bound
am = zeros(N,1); am_old = am; % active set for lower bound

% continuation in Moreau-Yosida-penalty gamma
for g = 10.^(0:12)
    if dispopt
        fprintf('Compute minimizer for gamma=%1.0e\n',g);
    end
    
    % semi-smooth Newton method
    for iter = 1:maxit
        Lx =  g*Kt*((ap+am).*yd);
        Lc =  g*(am - ap)'*yd;
        
        Lxx = aI + g*Kt*(spdiags(ap+am,0,N,N)*K);
        Lxc = g*Kt*(am-ap);
        Lcx = g*(am-ap)'*K;
        Lcc = N + g*sum(ap+am);
        
        H = [Lxx Lxc ; Lcx Lcc ];
        F = [Lx ; Lc];
        
        % Newton step
        dx = H\F;
        x  = dx(1:N);
        c  = dx(end);
        
        % compute active sets
        res = K*x-yd;             % residual
        ap  = (res>c);            % active set for upper constraint
        am  = (res<-c);           % active set for lower constraint
        
        % terminate if active sets coincide
        update = nnz((ap-ap_old)|(am-am_old));
        if dispopt && iter>0
            figure(3), plot(1:N,xe,'k',1:N,x,'r');xlabel('ssn'); drawnow
            fprintf('SSN iteration %d: %d points changed in active set\n',iter,update);
        end
        if update == 0
            break
        end
        
        ap_old = ap;
        am_old = am;
    end
    
    % terminate if minimizer for previous gamma feasible
    if iter == 1
        break
    end
end
