function [x,c] = linffitting_solve2d(AA,yd,alpha,A,opts)
% [y,c] = LINFFITTING_SOLVE2D(AA,yd,alpha,xe,opts)
% semi-smooth Newton method for L^\inf-fitting-L^2-regularization using a
%   path-following method (inverse source problem in 2d)
% Input:  AA     -- differential operator A^T*A
%         yd     -- given data
%         alpha  -- regularization parameter for primal problem
%         A      -- differential operator
%         opts   -- options for semismooth Newton method (see below)
% Output: x      -- Linf reconstruction
%         c      -- Linf bound norm of residual
%
% Christian Clason(christian.clason@uni-graz.at)
% February 9, 2012

%% Set parameters
NN = length(yd);

% Default parameters
if nargin <= 4 || isempty(opts)
    maxit   = 20;              % maximum number of iterations
    dispopt = 0;               % plot iterates in SSN iteration
else
    maxit   = opts.maxit;      % maximum number of iterations
    dispopt = opts.dispopt;    % plot iterates in SSN iteration    
end

% precompute parts of Newton derivative
aAA = alpha*AA;

% initialize iterates
ap = zeros(NN,1); ap_old = ap; % active set for upper bound
am = zeros(NN,1); am_old = am; % active set for lower bound

% continuation in Moreau-Yosida-penalty gamma
for g = 10.^(0:12)
    if dispopt
        fprintf('Compute minimizer for gamma=%1.0e\n',g);
    end
    
    % semi-smooth Newton method
    for iter = 1:maxit
        C = aAA+g*spdiags(ap+am,0,NN,NN);
        H = [C, g*(am-ap); g*(am-ap)', NN+g*sum(ap+am)];
        F = [g*(ap+am).*yd;g*(am-ap)'*yd];
        
        % Newton step
        dx = H\F;
        y  = dx(1:end-1);
        c  = dx(end);
        
        % compute active sets
        ap = (y-yd>c);          % active set for upper constraint
        am = (y-yd<-c);         % active set for lower constraint
        
        % terminate if active sets coincide
        update = nnz((ap-ap_old) | (am-am_old));
        if dispopt && iter>0
            figure(999), surf(reshape(A*y,sqrt(NN),sqrt(NN)));
            shading interp; camlight headlight; lighting phong; drawnow;
            fprintf('SSN iteration %d: %d points changed in active set\n', iter,update);
        end
        if update == 0
            break
        end
        
        ap_old = ap;
        am_old = am;
    end
    
    % terminate if minimizer for previous gamma feasible
    if iter == 1
        break;
    end
end

% computed L^\inf reconstruction
x = A*y;
