% LINFFITTING_TEST
% test script for L^\inf data fitting using automatic parameter choice and
% semismooth Newton method with continuation.
%
% Christian Clason(christian.clason@uni-graz.at)
% February 9, 2012

clear all; close all
%% Parameters
n     = 300;        % problem size
type  = 'uniform';  % type of noise ('uniform', 'quantized', 'checker')
d_mag = 0.1;        % magnitude of noise (number of bins for 'uniform')

% ... for balancing principle
alpha0  = 1e-1;     % initial guess for parameter alpha
sigma   = 0.008;    % balancing parameter
maxit   = 20;       % maximum number of fixed point iterations
tol     = 1e-3;     % terminate if relative change in alpha less
dispopt = 1;        % output iterates in fixed point iteration

% ... for semismooth Newton method
ssn_opt.maxit     = 20;       % maximum number of semismooth Newton steps
ssn_opt.dispopt   = 0;        % plot iterates in SSN iteration

%% Data
% operator K, exact data ye, reference solution xe
% heat from Regularization Tools: http://www2.imm.dtu.dk/~pch/Regutools/
[K,ye,xe] = heat(n);

% add noise to data
switch type
    case 'uniform'                      % uniform noise
        nmax  = d_mag*max(abs(ye));
        noise = nmax*2*(rand(n,1)-0.5);
        yd    = ye + noise;
    case 'quantized'                    % quantization noise
        nmax  = d_mag*(max(ye)-min(ye));
        yd    = round(ye/nmax)*nmax;
    case 'checker'                      % checkerboard noise
        noise = (-1).^(1:n);
        yd    = ye + d_mag*max(ye)*noise(:);
end

% compute exact noise level
delta = norm(yd-ye,inf);

% show noisy data
xx = linspace(0,1,n);
figure(1), plot(xx,ye,'k',xx,yd,'r');legend('exact','noisy');

%% Balancing principle loop
t_b = tic;
alpha = alpha0;
for iter = 1:maxit
    aold  = alpha;
    
    % compute x_alpha
    t_s   = tic;
    [x,c] = linffitting_solve(K,yd,alpha,xe,ssn_opt);
    t_ssn = toc(t_s);
    
    % update alpha
    alpha = sigma*2*n*c/norm(x)^2;
    err   = norm(x-xe)/sqrt(n);    % L^2 norm of reconstruction error
    
    % output
    if dispopt
        figure(99),plot(xx,x,'r',xx,xe,'k');drawnow
        display(sprintf('Iteration %d: t = %g seconds, alpha = %g, err= %g',...
            iter,t_ssn,alpha,err));
    end
    
    % terminate if relative change in alpha too small
    if abs(aold-alpha)<tol*aold
        break;
    end
end
t_balance = toc(t_b);

display(sprintf(['\nNoise level: delta = %e, estimated: c = %e\n' ...
    'Balancing principle returned alpha = %0.3e after %d iterations (%0.2f seconds).\n' ...
    'Final reconstruction error: %0.3e, reconstruction time: %0.2f seconds.\n'],...
    delta,c,alpha,iter,t_balance,err,t_ssn));

%% Compare with L2 fitting
% precompute operators
KtK = K'*K; I = speye(n); Ky = K'*yd;

% compute best L2 reconstruction
alrange = linspace(alpha/100,alpha,101);
err_opt = 1e99;
for altest = alrange
    xtest = (KtK+altest*I)\(Ky);
    err2 = norm(xtest-xe)/sqrt(n);
    if err2 < err_opt
        err_opt = err2;
        alpha2 = altest;
        x2     = xtest;
    end
end

display(sprintf('L2 reconstruction error (alpha=%0.3e): %0.3e',alpha2,err_opt));

%% Plot reconstructions

figure(3),plot(xx,x,'r',xx,x2,'b--',xx,xe,'k');
title('Reconstruction');
legend(['L^\infty \alpha=' num2str(alpha,'%0.2e')], ['L^2 \alpha=' num2str(alpha2,'%0.2e')],'exact');
