% LINFFITTING_TEST2D
% test script for L^\inf data fitting using automatic parameter choice and
% semismooth Newton method with continuation (inverse source problem in 2D).
%
% Christian Clason(christian.clason@uni-graz.at)
% February 9, 2012

clear all; close all
%% parameters
N     = 64; n = N^2;  % grid size
type  = 'checker';    % type of noise ('uniform', 'quantized', 'checker')
d_mag = 0.9;          % magnitude of noise (1/number of bins for 'uniform')

% ... for balancing principle
alpha0  = 1e-1;       % initial guess for parameter alpha
sigma   = 0.008;      % balancing parameter
maxit   = 20;         % maximum number of fixed point iterations
tol     = 1e-3;       % terminate if relative change in alpha less
dispopt = 1;          % output iterates in fixed point iteration

% ... for semismooth Newton method
ssn_opt.maxit   = 20; % maximum number of semismooth Newton steps
ssn_opt.dispopt = 0;  % plot iterates in SSN iteration

%% Data
% coefficients of differential operator
nu = 1;               % diffusion
b  = -[2,0];          % convection

% setup grid on [-1,1]^2
[xx,yy] = meshgrid(linspace(-1,1,N)); h = 2/N;

% setup operators
Dxx = gallery('poisson',N)*1/h^2; % -Laplace
e   = ones(N,1);
Dx  = 1/h*spdiags([-e e],-1:0,N,N);
Ix  = speye(N,N);

A   = nu*Dxx+b(1)*kron(Ix,Dx)+b(2)*kron(Dx,Ix); AA = A'*A;
xe  = 1.0*(abs(xx)<.33).*(abs(yy)<.33);  % exact solution
ye  = A\xe(:);                           % exact data

% add noise to data
switch type
    case 'uniform'                      % uniform noise
        nmax  = d_mag*max(abs(ye));
        noise = nmax*2*(rand(n,1)-0.5);
        yd    = ye + noise;
    case 'quantized'                    % quantization noise
        nmax  = d_mag*(max(ye)-min(ye));
        yd    = round(ye/nmax)*nmax;
    case 'checker'                      % checkerboard noise
        noise = (-1).^((1:N)'*(1:N));
        yd    = ye + d_mag*max(ye)*noise(:);
end

% compute exact noise level
delta = norm(yd-ye,inf);

% show the true solution and exact, noisy data
figure(1); surf(xx,yy,reshape(xe,N,N)); title('exact solution');
shading interp; camlight headlight; lighting phong; drawnow;
figure(2); surf(xx,yy,reshape(ye,N,N)); title('exact data');
shading interp; camlight headlight; lighting phong; drawnow;
figure(3); surf(xx,yy,reshape(yd,N,N)); title('noisy data');
shading interp; camlight headlight; lighting phong; drawnow;

%% Balancing principle loop
t_b = tic;

alpha = alpha0;
for iter = 1:20
    alpha1 = alpha;
    
    % compute x_\alpha
    t_s   = tic;
    [x,c] = linffitting_solve2d(AA,yd,alpha,A,ssn_opt);
    t_ssn = toc(t_s);
    
    % update alpha
    alpha = sigma*c/(norm(x)^2/(2*n));
    err   = norm(x-xe(:))/sqrt(n);
    
    if dispopt
        figure(99);  surf(xx,yy,reshape(x,N,N));
        shading interp; camlight headlight; lighting phong; axis tight; drawnow;
        display(sprintf('Iteration %d: t = %g seconds, alpha = %g, err= %g',...
            iter,t_ssn,alpha,err));
    end
    
    % terminate if relative change in alpha too small
    if (abs(alpha-alpha1)/alpha<tol)
        break;
    end
end
t_balance = toc(t_b);

display(sprintf(['\nNoise level: delta = %e, estimated: c= %e\n' ...
    'Balancing principle returned alpha = %0.3e after %d iterations (%0.2f seconds).\n' ...
    'Final reconstruction error: %0.3e, reconstruction time: %0.2f seconds.'],...
    delta,c,alpha,iter,t_balance,err,t_ssn));

%% compare with L2 fitting
I = speye(n);

alrange = linspace(alpha/100,alpha,101);
err_opt = 1e99;
for altest = alrange
    ytest = (altest*AA+I)\yd;
    xtest = A*ytest;
    err2  = norm(xtest-xe(:))/N;
    if err2 < err_opt
        err_opt = err2;
        alpha2  = altest;
        x2      = xtest;
    end
end

display(sprintf('L2 reconstruction error (alpha=%0.3e): %0.3e',alpha2,err_opt));

%% show reconstruction
figure(4); surf(xx,yy,reshape(x,N,N)); title(['L1 reconstruction (\alpha=' num2str(alpha,'%0.2e') ')']);
shading interp; camlight headlight; lighting phong; axis tight; drawnow;
figure(5); surf(xx,yy,reshape(x2,N,N)); title(['L2 reconstruction (\alpha=' num2str(alpha2,'%0.2e') ')']);
shading interp; camlight headlight; lighting phong; axis tight; drawnow;
